/****************************************************************************
** Meta object code from reading C++ file 'qcanbusframe.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.2)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtserialbus/src/serialbus/qcanbusframe.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qcanbusframe.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.2. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN12QCanBusFrameE_t {};
} // unnamed namespace

template <> constexpr inline auto QCanBusFrame::qt_create_metaobjectdata<qt_meta_tag_ZN12QCanBusFrameE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QCanBusFrame",
        "FrameType",
        "UnknownFrame",
        "DataFrame",
        "ErrorFrame",
        "RemoteRequestFrame",
        "InvalidFrame",
        "FrameError",
        "NoError",
        "TransmissionTimeoutError",
        "LostArbitrationError",
        "ControllerError",
        "ProtocolViolationError",
        "TransceiverError",
        "MissingAcknowledgmentError",
        "BusOffError",
        "BusError",
        "ControllerRestartError",
        "UnknownError",
        "AnyError",
        "FrameErrors"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'FrameType'
        QtMocHelpers::EnumData<enum FrameType>(1, 1, QMC::EnumFlags{}).add({
            {    2, FrameType::UnknownFrame },
            {    3, FrameType::DataFrame },
            {    4, FrameType::ErrorFrame },
            {    5, FrameType::RemoteRequestFrame },
            {    6, FrameType::InvalidFrame },
        }),
        // enum 'FrameError'
        QtMocHelpers::EnumData<enum FrameError>(7, 7, QMC::EnumFlags{}).add({
            {    8, FrameError::NoError },
            {    9, FrameError::TransmissionTimeoutError },
            {   10, FrameError::LostArbitrationError },
            {   11, FrameError::ControllerError },
            {   12, FrameError::ProtocolViolationError },
            {   13, FrameError::TransceiverError },
            {   14, FrameError::MissingAcknowledgmentError },
            {   15, FrameError::BusOffError },
            {   16, FrameError::BusError },
            {   17, FrameError::ControllerRestartError },
            {   18, FrameError::UnknownError },
            {   19, FrameError::AnyError },
        }),
        // flag 'FrameErrors'
        QtMocHelpers::EnumData<FrameErrors>(20, 7, QMC::EnumIsFlag).add({
            {    8, FrameError::NoError },
            {    9, FrameError::TransmissionTimeoutError },
            {   10, FrameError::LostArbitrationError },
            {   11, FrameError::ControllerError },
            {   12, FrameError::ProtocolViolationError },
            {   13, FrameError::TransceiverError },
            {   14, FrameError::MissingAcknowledgmentError },
            {   15, FrameError::BusOffError },
            {   16, FrameError::BusError },
            {   17, FrameError::ControllerRestartError },
            {   18, FrameError::UnknownError },
            {   19, FrameError::AnyError },
        }),
    };
    return QtMocHelpers::metaObjectData<QCanBusFrame, qt_meta_tag_ZN12QCanBusFrameE_t>(QMC::PropertyAccessInStaticMetaCall, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QCanBusFrame::staticMetaObject = { {
    nullptr,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QCanBusFrameE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12QCanBusFrameE_t>.data,
    nullptr,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN12QCanBusFrameE_t>.metaTypes,
    nullptr
} };

QT_WARNING_POP
