/****************************************************************************
** Meta object code from reading C++ file 'qcanbusdevice.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.2)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../qtserialbus/src/serialbus/qcanbusdevice.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'qcanbusdevice.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.2. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN13QCanBusDeviceE_t {};
} // unnamed namespace

template <> constexpr inline auto QCanBusDevice::qt_create_metaobjectdata<qt_meta_tag_ZN13QCanBusDeviceE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QCanBusDevice",
        "errorOccurred",
        "",
        "QCanBusDevice::CanBusError",
        "framesReceived",
        "framesWritten",
        "framesCount",
        "stateChanged",
        "QCanBusDevice::CanBusDeviceState",
        "state",
        "CanBusError",
        "NoError",
        "ReadError",
        "WriteError",
        "ConnectionError",
        "ConfigurationError",
        "UnknownError",
        "OperationError",
        "TimeoutError",
        "CanBusDeviceState",
        "UnconnectedState",
        "ConnectingState",
        "ConnectedState",
        "ClosingState",
        "CanBusStatus",
        "Unknown",
        "Good",
        "Warning",
        "Error",
        "BusOff",
        "ConfigurationKey",
        "RawFilterKey",
        "ErrorFilterKey",
        "LoopbackKey",
        "ReceiveOwnKey",
        "BitRateKey",
        "CanFdKey",
        "DataBitRateKey",
        "ProtocolKey",
        "UserKey"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'errorOccurred'
        QtMocHelpers::SignalData<void(QCanBusDevice::CanBusError)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'framesReceived'
        QtMocHelpers::SignalData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'framesWritten'
        QtMocHelpers::SignalData<void(qint64)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 6 },
        }}),
        // Signal 'stateChanged'
        QtMocHelpers::SignalData<void(QCanBusDevice::CanBusDeviceState)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 9 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
        // enum 'CanBusError'
        QtMocHelpers::EnumData<enum CanBusError>(10, 10, QMC::EnumFlags{}).add({
            {   11, CanBusError::NoError },
            {   12, CanBusError::ReadError },
            {   13, CanBusError::WriteError },
            {   14, CanBusError::ConnectionError },
            {   15, CanBusError::ConfigurationError },
            {   16, CanBusError::UnknownError },
            {   17, CanBusError::OperationError },
            {   18, CanBusError::TimeoutError },
        }),
        // enum 'CanBusDeviceState'
        QtMocHelpers::EnumData<enum CanBusDeviceState>(19, 19, QMC::EnumFlags{}).add({
            {   20, CanBusDeviceState::UnconnectedState },
            {   21, CanBusDeviceState::ConnectingState },
            {   22, CanBusDeviceState::ConnectedState },
            {   23, CanBusDeviceState::ClosingState },
        }),
        // enum 'CanBusStatus'
        QtMocHelpers::EnumData<enum CanBusStatus>(24, 24, QMC::EnumIsScoped).add({
            {   25, CanBusStatus::Unknown },
            {   26, CanBusStatus::Good },
            {   27, CanBusStatus::Warning },
            {   28, CanBusStatus::Error },
            {   29, CanBusStatus::BusOff },
        }),
        // enum 'ConfigurationKey'
        QtMocHelpers::EnumData<enum ConfigurationKey>(30, 30, QMC::EnumFlags{}).add({
            {   31, ConfigurationKey::RawFilterKey },
            {   32, ConfigurationKey::ErrorFilterKey },
            {   33, ConfigurationKey::LoopbackKey },
            {   34, ConfigurationKey::ReceiveOwnKey },
            {   35, ConfigurationKey::BitRateKey },
            {   36, ConfigurationKey::CanFdKey },
            {   37, ConfigurationKey::DataBitRateKey },
            {   38, ConfigurationKey::ProtocolKey },
            {   39, ConfigurationKey::UserKey },
        }),
    };
    return QtMocHelpers::metaObjectData<QCanBusDevice, qt_meta_tag_ZN13QCanBusDeviceE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QCanBusDevice::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QCanBusDeviceE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QCanBusDeviceE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13QCanBusDeviceE_t>.metaTypes,
    nullptr
} };

void QCanBusDevice::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QCanBusDevice *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->errorOccurred((*reinterpret_cast<std::add_pointer_t<QCanBusDevice::CanBusError>>(_a[1]))); break;
        case 1: _t->framesReceived(); break;
        case 2: _t->framesWritten((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 3: _t->stateChanged((*reinterpret_cast<std::add_pointer_t<QCanBusDevice::CanBusDeviceState>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (QCanBusDevice::*)(QCanBusDevice::CanBusError )>(_a, &QCanBusDevice::errorOccurred, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (QCanBusDevice::*)()>(_a, &QCanBusDevice::framesReceived, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (QCanBusDevice::*)(qint64 )>(_a, &QCanBusDevice::framesWritten, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (QCanBusDevice::*)(QCanBusDevice::CanBusDeviceState )>(_a, &QCanBusDevice::stateChanged, 3))
            return;
    }
}

const QMetaObject *QCanBusDevice::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QCanBusDevice::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13QCanBusDeviceE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int QCanBusDevice::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 4)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 4;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 4)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 4;
    }
    return _id;
}

// SIGNAL 0
void QCanBusDevice::errorOccurred(QCanBusDevice::CanBusError _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void QCanBusDevice::framesReceived()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void QCanBusDevice::framesWritten(qint64 _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void QCanBusDevice::stateChanged(QCanBusDevice::CanBusDeviceState _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}
QT_WARNING_POP
