/****************************************************************************
** Meta object code from reading C++ file 'passthrucanio.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.2)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../../qtserialbus/src/plugins/canbus/passthrucan/passthrucanio.h"
#include <QtCore/qmetatype.h>
#include <QtCore/QList>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'passthrucanio.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.2. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN13PassThruCanIOE_t {};
} // unnamed namespace

template <> constexpr inline auto PassThruCanIO::qt_create_metaobjectdata<qt_meta_tag_ZN13PassThruCanIOE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "PassThruCanIO",
        "errorOccurred",
        "",
        "description",
        "QCanBusDevice::CanBusError",
        "error",
        "messagesReceived",
        "QList<QCanBusFrame>",
        "frames",
        "messagesSent",
        "count",
        "openFinished",
        "success",
        "closeFinished",
        "open",
        "library",
        "subDev",
        "bitRate",
        "close",
        "applyConfig",
        "QCanBusDevice::ConfigurationKey",
        "key",
        "QVariant",
        "value",
        "listen"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'errorOccurred'
        QtMocHelpers::SignalData<void(const QString &, QCanBusDevice::CanBusError)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 3 }, { 0x80000000 | 4, 5 },
        }}),
        // Signal 'messagesReceived'
        QtMocHelpers::SignalData<void(QList<QCanBusFrame>)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Signal 'messagesSent'
        QtMocHelpers::SignalData<void(qint64)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 10 },
        }}),
        // Signal 'openFinished'
        QtMocHelpers::SignalData<void(bool)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 12 },
        }}),
        // Signal 'closeFinished'
        QtMocHelpers::SignalData<void()>(13, 2, QMC::AccessPublic, QMetaType::Void),
        // Method 'open'
        QtMocHelpers::MethodData<void(const QString &, const QByteArray &, uint)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 15 }, { QMetaType::QByteArray, 16 }, { QMetaType::UInt, 17 },
        }}),
        // Method 'close'
        QtMocHelpers::MethodData<void()>(18, 2, QMC::AccessPublic, QMetaType::Void),
        // Method 'applyConfig'
        QtMocHelpers::MethodData<void(QCanBusDevice::ConfigurationKey, const QVariant &)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 20, 21 }, { 0x80000000 | 22, 23 },
        }}),
        // Method 'listen'
        QtMocHelpers::MethodData<void()>(24, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PassThruCanIO, qt_meta_tag_ZN13PassThruCanIOE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject PassThruCanIO::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13PassThruCanIOE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13PassThruCanIOE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN13PassThruCanIOE_t>.metaTypes,
    nullptr
} };

void PassThruCanIO::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PassThruCanIO *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->errorOccurred((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QCanBusDevice::CanBusError>>(_a[2]))); break;
        case 1: _t->messagesReceived((*reinterpret_cast<std::add_pointer_t<QList<QCanBusFrame>>>(_a[1]))); break;
        case 2: _t->messagesSent((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 3: _t->openFinished((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 4: _t->closeFinished(); break;
        case 5: _t->open((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QByteArray>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<uint>>(_a[3]))); break;
        case 6: _t->close(); break;
        case 7: _t->applyConfig((*reinterpret_cast<std::add_pointer_t<QCanBusDevice::ConfigurationKey>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVariant>>(_a[2]))); break;
        case 8: _t->listen(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PassThruCanIO::*)(const QString & , QCanBusDevice::CanBusError )>(_a, &PassThruCanIO::errorOccurred, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PassThruCanIO::*)(QList<QCanBusFrame> )>(_a, &PassThruCanIO::messagesReceived, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PassThruCanIO::*)(qint64 )>(_a, &PassThruCanIO::messagesSent, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PassThruCanIO::*)(bool )>(_a, &PassThruCanIO::openFinished, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (PassThruCanIO::*)()>(_a, &PassThruCanIO::closeFinished, 4))
            return;
    }
}

const QMetaObject *PassThruCanIO::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *PassThruCanIO::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN13PassThruCanIOE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int PassThruCanIO::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 9)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 9;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 9)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 9;
    }
    return _id;
}

// SIGNAL 0
void PassThruCanIO::errorOccurred(const QString & _t1, QCanBusDevice::CanBusError _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2);
}

// SIGNAL 1
void PassThruCanIO::messagesReceived(QList<QCanBusFrame> _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void PassThruCanIO::messagesSent(qint64 _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void PassThruCanIO::openFinished(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void PassThruCanIO::closeFinished()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}
QT_WARNING_POP
